package org.bitbucket.bradleysmithllc.etlunit.feature.execute;

import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.execute.json.java_executor.execute.ExecuteHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.execute.json.java_executor.execute.ExecuteRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.execute.json.java_executor.execute.SystemProperties;
import org.bitbucket.bradleysmithllc.etlunit.feature.logging.LogFileManager;
import org.bitbucket.bradleysmithllc.etlunit.io.JavaForker;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

public class JavaExecutor extends NullClassListener implements ExecuteHandler {
	private RuntimeSupport runtimeSupport;
	private LogFileManager logFileManager;

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport support)
	{
		runtimeSupport = support;
	}

	@Inject
	public void receiveLogFileManager(LogFileManager manager)
	{
		logFileManager = manager;
	}

	@Override
	public action_code execute(ExecuteRequest request, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		String className = request.getJavaClass();

		try
		{
			Class cl = Thread.currentThread().getContextClassLoader().loadClass(className);

			Method mth = cl.getDeclaredMethod("main", new Class [] {String[].class});

			int mods = mth.getModifiers();

			if (!Modifier.isPublic(mods))
			{
				throw new TestExecutionError("Java class has main method, but it isn't public: " + mth, ExecuteConstants.ERR_JAVA_MAIN_NOT_PUBLIC);
			}

			if (!Modifier.isStatic(mods))
			{
				throw new TestExecutionError("Java class has public main method, but it isn't static: " + mth, ExecuteConstants.ERR_JAVA_MAIN_NOT_STATIC);
			}

			List<String> requestArguments = request.getArguments();

			String [] args;

			JavaForker javaForker = new JavaForker(runtimeSupport);

			// use the temp folder as the working dir
			javaForker.setWorkingDirectory(runtimeSupport.getTempDirectory());

			javaForker.setMainClassName(className);

			// adjust the classpath if running in maven
			String property = System.getProperty("etlunit.maven.plugin.classpath");
			if (property != null)
			{
				String [] pathElements = property.split(Pattern.quote(File.pathSeparator));

				for (String pathElement : pathElements)
				{
					javaForker.addClasspathEntry(new File(pathElement));
				}
			}

			if (requestArguments != null && requestArguments.size() != 0)
			{
				args = new String[requestArguments.size()];

				for (int i = 0; i < args.length; i++)
				{
					String argument = requestArguments.get(i);
					javaForker.addArgument(argument);

					args[i] = argument;
				}
			}

			// set a debug token in the system properties to identify this test
			javaForker.addSystemProperty("etlunit.active", "true");
			javaForker.addSystemProperty("etlunit.test.class", testOperation.getTestClass().getQualifiedName());
			javaForker.addSystemProperty("etlunit.test.method", testOperation.getTestMethod().getName());
			javaForker.addSystemProperty("etlunit.test.method.qualified", testOperation.getTestMethod().getQualifiedName());
			javaForker.addSystemProperty("etlunit.test.operation", testOperation.getOperationName());
			javaForker.addSystemProperty("etlunit.test.operation.ordinal", String.valueOf(testOperation.getOrdinal()));
			javaForker.addSystemProperty("etlunit.test.operation.qualified", testOperation.getQualifiedName());

			SystemProperties properties = request.getSystemProperties();

			if (properties != null)
			{
				Map<String,String> additionalProperties = properties.getAdditionalProperties();

				if (additionalProperties != null)
				{
					System.getProperties().putAll(additionalProperties);
					javaForker.addSystemProperties(additionalProperties);
				}
			}

			// check for the ext dirs property so we can propogate it
			String extDirs = System.getProperty("java.ext.dirs");

			if (extDirs != null)
			{
				// convert this into an absolute path so that it resolves properly
				javaForker.addSystemProperty("java.ext.dirs", new File(extDirs).getAbsolutePath());
			}

			String runId = className + "_" + testOperation.getQualifiedName() + ".out";

			if (request.getRunId() != null)
			{
				runId = request.getRunId() + ".out";
			}

			File file = runtimeSupport.createGeneratedSourceFile("java-executor", runId);

			javaForker.setOutput(file);

			ProcessFacade proc = javaForker.startProcess();
			proc.waitForCompletion();

			// attach standard out to the log files
			if (file.exists())
			{
				logFileManager.addLogFile(testOperation, file, "standard-out");
			}

			if (request.getExitCode() != null)
			{
				int ret = proc.getCompletionCode();

				int exp = request.getExitCode().intValue();
				if (ret != exp)
				{
					throw new TestExecutionError("Unexpected process exit code.  [" + exp + "] expected, [" + ret + "] observed.", ExecuteConstants.ERR_JAVA_EXECUTION_ERROR_BAD_RETURN_CODE);
				}
			}
		}
		catch (ClassNotFoundException e)
		{
			throw new TestExecutionError("Java class not found: " + className, ExecuteConstants.ERR_JAVA_CLASS_NOT_FOUND, e);
		}
		catch (NoSuchMethodException e)
		{
			throw new TestExecutionError("Java class found, but it does not have a main method: " + className, ExecuteConstants.ERR_JAVA_MAIN_METHOD_NOT_FOUND, e);
		} catch (IOException e) {
			throw new TestExecutionError("Java execution failure: " + className, ExecuteConstants.ERR_JAVA_EXECUTION_ERROR, e);
		}

		return action_code.handled;
	}
}
